"""Declarative, dynamic option parsing."""

# fmt: off
__version__ = "3.0.0"
__all__ = (
    # Options & arguments
    # -------------------
    "take_opt",
    "argv",

    # Error handling
    # --------------
    "show_errors",
    "expect",
    "fail",

    # Exceptions
    # ----------
    "ArgsError",
    "MissingOption",
    "TransformError",
    "OptionError",
)
# fmt: on

from lethargy.errors import ArgsError, MissingOption, OptionError, TransformError
from lethargy.option import take_opt
from lethargy.util import argv, expect, fail, show_errors



"""Defines the Opt class (main interface)."""

from lethargy.errors import ArgsError, MissingOption, TransformError
from lethargy.util import argv, falsylist, identity, is_greedy, tryposixname


class Option:
    """Define an option to take it from a list of arguments."""

    def __init__(self, name, number=0, tfm=None):
        self.names = set(map(tryposixname, [name] if isinstance(name, str) else name))
        self.argc = number  # Invalid values are handled by the setter.
        self.tfm = tfm or identity

    def __str__(self):
        if not self.names:
            return ""

        names = "|".join(sorted(sorted(self.names), key=len))

        if not isinstance(self.tfm, type):
            metavar = "value"
        else:
            metavar = self.tfm.__name__.lower()

        if is_greedy(self.argc):
            args = f"[{metavar}]..."
        elif self.argc > 0:
            args = " ".join([f"<{metavar}>"] * self.argc)
        else:
            return names

        return f"{names} {args}"

    def __repr__(self):
        repr_str = ""

        # Option(<names>)
        qname = type(self).__qualname__
        mapped = [repr(name) for name in self.names]
        names = ", ".join(mapped)
        repr_str += f"{qname}({names})"

        # [.takes(<n>[, <tfm>])]
        # This whole thing is optional. If there's nothing
        # to show, it won't be in the final string.
        if self.argc:
            takes = [self.argc]
            if self.tfm is not identity:
                if isinstance(self.tfm, type):
                    takes.append(self.tfm.__name__)
                else:
                    takes.append(repr(self.tfm))
            repr_str += ".takes({})".format(", ".join(map(str, takes)))

        # at <ID>
        repr_str += f" at {hex(id(self))}"

        return f"<{repr_str}>"

    @property
    def argc(self):
        """Get the number of arguments this option takes."""
        return self._argc

    @argc.setter
    def argc(self, value):
        if not is_greedy(value) and value < 0:
            msg = f"The number of arguments ({value}) must be >1 or greedy (``...``)"
            raise ValueError(msg)
        self._argc = value

    def find_in(self, args):
        """Search args for this option and return an index if it's found."""
        for name in self.names:
            try:
                return args.index(name)
            except ValueError:
                continue
        return None

    def take_flag(self, args=argv, *, mut=True):
        """Get a bool indicating whether the option was present in the arguments."""
        index = self.find_in(args)

        if index is None:
            return False

        if mut:
            del args[index]

        return True

    def take_args(self, args=argv, *, required=False, mut=True):
        """Get the values of this option."""
        argc = self.argc

        if not argc:
            msg = f"'{self}' takes no arguments (did you mean to use `take_flag`?)"
            raise RuntimeError(msg)

        # Is this option in the list?
        index = self.find_in(args)

        # Return early if the option isn't present.
        if index is None:
            if required:
                msg = f"Missing required option '{self}'"
                raise MissingOption(msg)

            if is_greedy(argc):
                return falsylist()

            if argc != 1:
                return falsylist([None] * argc)

            return None

        # Start index is now set, find the index of the *final* value.
        if is_greedy(argc):
            end_idx = None
        else:
            # Start index is the option name, add 1 to compensate.
            end_idx = index + argc + 1

            # Fail fast if the option expects more arguments than it has.
            if end_idx > len(args):
                # Highest index (length - 1) minus this option's index.
                number_found = len(args) - 1 - index
                n = number_found or "none"
                s = "s" if argc != 1 else ""
                msg = f"Expected {argc} argument{s} for option '{self}', but found {n}"
                if number_found:
                    given = ", ".join(map(repr, args[index + 1 : end_idx]))
                    msg += f" ({given})"
                raise ArgsError(msg)

        # Get the list of values starting from the first value to the option.
        taken = args[index + 1 : end_idx]

        # Remove the option and its associated values from the list.
        if mut:
            del args[index:end_idx]

        # Single return value keeps the unpacking usage pattern consistent.
        if argc == 1:
            return self.transform(taken[0])

        # Return a list of transformed values.
        return [self.transform(x) for x in taken]

    def transform(self, value):
        """Call _tfm on a string and return the result, or raise an exception union."""
        try:
            return self.tfm(value)

        except Exception as exc:
            message = f"Option '{self}' received an invalid value: {value!r}"

            # The exception needs to be a subclass of both the raised exception
            # and TransformError. This allows manually handling specific
            # exception types, _and_ automatically handling all exceptions that
            # get raised during transformation.
            name = f"TransformError[{type(exc).__name__}]"
            bases = (TransformError, type(exc))
            new_exc = type(name, bases, {})

            raise new_exc(message) from exc


def take_opt(name, number=None, call=None, *, args=argv, required=False, mut=True):
    """Quickly take an option as flag, or with some arguments if also given a number."""
    if not number:
        return Option(name).take_flag(args, mut=mut)

    opt = Option(name, number, call)

    return opt.take_args(args, required=required, mut=mut)



"""Functions and values, independent of other modules."""

import sys
from contextlib import contextmanager

from lethargy.errors import OptionError, TransformError

# Lethargy provides its own argv so you don't have to import sys or worry
# about mutating the original.
argv = sys.argv.copy()


def tryposixname(text):
    """Get a POSIX-style name, or strip if the first character isn't alphanumeric."""
    stripped = str(text).strip()

    # Assume it's been pre-formatted if it starts with something that's not
    # a letter or number.
    if not stripped[:1].isalnum():
        return stripped

    name = "-".join(stripped.split())

    chars = len(name)

    if chars > 1:
        return f"--{name}"
    if chars == 1:
        return f"-{name}"

    raise ValueError("Cannot make an option name from an empty string.")


def is_greedy(value):
    """Return a boolean representing whether a given value is "greedy"."""
    return value is ...


def identity(a):
    """Get the same output as the input."""
    return a


def fail(message=None):
    """Print a message to stderr and exit with code 1."""
    if message:
        print(message, file=sys.stderr)
    sys.exit(1)


@contextmanager
def expect(*errors, reason=None):
    """Call `fail()` if any given errors are raised."""
    try:
        yield
    except errors as e:
        fail(reason or e)


def show_errors():
    """Expect errors from options and values, fail with a useful message."""
    return expect(OptionError, TransformError)


falsylist = type("falsylist", (list,), {"__bool__": lambda _: False})



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lethargy']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'lethargy',
    'version': '3.0.0',
    'description': 'A minimal library to make your option-parsing easier.',
    'long_description': '# Lethargy: Terse & tiny command-line option library\n\n**Lethargy was born out of frustration**. It gets out of your way as soon as possible to let you get on with the actual logic. No bullshit, no magic, no objects to understand, you just call a function.\n\nI write a lot of small scripts to get my job done faster, and manually working with options is a pain. Existing libraries are extremely verbose or just don\'t feel good to use. _Lethargy is designed to make writing scripts easier and faster, and to reduce effort needed to maintain them_.\n\n<!-- Note that the spaces here are U+2000 (\'\u2002\') EN QUAD -->\n<!--                 v                                  -->\n- **No boilerplate.**\u2002Headaches are directly proportional to lines of code.\n- **No bloat.**\u2002Small API surface area, very little to learn.\n- **No ambiguity.**\u2002Lethargy raises exceptions instead of getting your code into bad state.\n- **Clear errors.**\u2002Great error messages and context managers for dealing with them.\n- **Flexible.**\u2002You\'re not locked in to any styles or paradigms.\n\nLethargy is completely imperative and is **not** a framework. If you _are_ building a complete CLI or want automatic help commands, you\'re better off using **[Click]** — a fantastic, declarative CLI framework.\n\n[Click]: https://click.palletsprojects.com/en/7.x/\n\n## Installation\n\nYou can use pip to install lethargy. It\'s tiny and only depends on the standard library.\n\n```console\npip install lethargy\n```\n\n## Usage\n\n```python\nimport lethargy\n\n# Accepts the option \'--bytes <int>\'. Show the error nicely if it goes wrong.\nwith lethargy.show_errors():\n    n_bytes = lethargy.take_opt(\'bytes\', 1, int) or 8\n\n# Now the option and value have been removed from lethargy.argv\nwith lethargy.expect(IndexError, reason=\'Missing required argument: [DIR]\'):\n    directory = lethargy.argv[1]\n\n...\n```\n\n## Getting Started\n\nThis is both a tutorial and the documentation. All examples assume you\'ve got `import lethargy` at the top.\n\n###### FLAGS\n\n**Options can be flags.** `True` if present, `False` if not.\n\n```python\n# --debug\ndebug = lethargy.take_opt(\'debug\')\n\nprint(debug)\n```\n\n```console\n$ python example.py --debug\nTrue\n$ python example.py\nFalse\n```\n\n<br>\n\n###### NAMES\n\n**Options can have more than one name.** Instead of a string, use a list of strings. Names are case-sensitive.\n\n```python\n# -v|--verbose\nverbose = lethargy.take_opt([\'v\', \'verbose\'])\n\nprint(verbose)\n```\n\n```console\n$ python example.py -v\nTrue\n$ python example.py --verbose\nTrue\n```\n\n<table><tbody><tr><td>💡</td><td>\n<!-- <tip> -->\nNames are created automatically (POSIX style) if the given names start with a letter or number. Names like <code>\'-test\'</code> and <code>\'/f\'</code> are treated as literal because of the first character.\n<!-- </tip> -->\n</td></tr></tbody></table><br>\n\n###### ARGUMENTS\n\n**Options can take arguments, too.** They can take any amount, and values are **always** space-separated.\n\n```python\n# -o|--output <value>\noutput = lethargy.take_opt([\'o\', \'output\'], 1)\n\nprint(output)\n```\n\n```console\n$ python example.py -o out.txt\nout.txt\n$ python example.py\nNone\n```\n\n<table><tbody><tr><td>💡</td><td>\n<!-- <tip> -->\nIf there are fewer values than what the option takes, it\'ll raise <code>lethargy.ArgsError</code>. See <a href="#error-handling">Error Handling</a> for how to present error messages nicely.\n<!-- </tip> -->\n</td></tr></tbody></table><br>\n\n###### GREEDINESS\n\n**Options can be variadic (greedy).** Use `...` instead of a number to take every value following the option.\n\n```python\n# -i|--ignore [value]...\nignored = lethargy.take_opt([\'i\', \'ignore\'], ...)\n\nfor pattern in ignored:\n    print(pattern)\n```\n\n```console\n$ python example.py --ignore .git .vscode .DS_Store\n.git\n.vscode\n.DS_Store\n$ python example.py --ignore experiments\nexperiments\n$ python example.py\n$ ▏\n```\n\n<table><tbody><tr><td>💡</td><td>\n<!-- <tip> -->\nVariadic options are greedy and will take <b>every</b> argument that follows them, including values that look like other options. You should always try and take these last (<i>after</i> taking the fixed-count options).\n<!-- </tip> -->\n</td></tr></tbody></table><br>\n\n###### UNPACKING\n\n**Unpack multiple values into separate variables.** If the option wasn\'t present, they\'ll all be `None`.\n\n```python\n# --name <value> <value> <value>\nfirst, middle, last = lethargy.take_opt(\'name\', 3)\n\nprint(f\'Hi, {first}!\')\n```\n\n```console\n$ python example.py --name Dwight Kurt Schrute\nHi, Dwight!\n$ python example.py\nHi, None!\n```\n\n<br>\n\n###### DEFAULTS\n\n**Set sensible defaults.** Use the `or` keyword and your default value(s).\n\n```python\n# -h|--set-hours <value> <value>\nstart, finish = lethargy.take_opt([\'set hours\', \'h\'], 2) or \'9AM\', \'5PM\'\n\nprint(f\'Employee now works {start} to {finish}\')\n```\n\n```console\n$ python example.py\nEmployee works 9AM to 5PM\n$ python example.py --set-hours 8AM 4PM\nEmployee works 8AM to 4PM\n```\n\n<table><tbody><tr><td>💡</td><td>\n<!-- <tip> -->\nYou should use defaults unless your option explicitly sets <code>required=True</code>. You\'ll thank yourself when you need to change something 6 months from now!\n<!-- </tip> -->\n</td></tr></tbody></table><br>\n\n###### TYPES & CONVERSION\n\n**Convert your option\'s values.** Use a function or type as the final argument. Defaults aren\'t converted.\n\n```python\n# --date-ymd <int> <int> <int>\ny, m, d = lethargy.take_opt(\'date ymd\', 3, int) or 1970, 1, 1\n\nfrom datetime import datetime\ndate = datetime(y, m, d)\ndelta = datetime.today() - date\nprint(f\'it has been {delta.days} days since {date}\')\n```\n\n```console\n$ python example.py --date-ymd 1999 10 9\nit has been 7500 days since 1999-10-09 00:00:00\n```\n\n<br>\n\n###### ERROR HANDLING\n\n**Give clear error messages.** Lethargy makes this easy with simple context managers.\n\n```python\nwith lethargy.show_errors():\n    x, y = lethargy.take_opt([\'p\', \'pos\'], 2, int) or 0, 0\n```\n\n```console\n$ python example.py --pos 20\nExpected 2 arguments for option \'-p|--pos <int> <int>\', but found 1 (\'20\')\n$ python example.py -p 20, 0\nOption \'-p|--pos <int> <int>\' received an invalid value: \'20,\'\n```\n\n<details>\n<summary align="right">Learn more about handling errors</summary>\n<br>\n\nUse `fail()` to exit with status code 1. You can optionally give it a message.\n\nLethargy provides two context managers for easier error handling. These share similar behaviour, but are separate to make intent clearer.\n\n> <i>with</i> <code><i>lethargy.</i><b>expect(</b><i>*errors: Exception, reason: Optional[str] = None</i><b>)</b></code>\n\nWhen one of the given exceptions is raised, it calls `fail()` to exit and print the message.\n\n> <i>with</i> <code><i>lethargy.</i><b>show_errors()</b></code>\n\nSame behaviour as `expect`, but specifically for handling options. Exceptions raised during value conversions will also be caught by `show_errors()`, with a useful message.\n\n<table><tbody><tr><td>💡</td><td>\n<!-- <tip> -->\nYou can access the original exception that caused a <code>TransformError</code> with the <code>__cause__</code> attribute (see the Python <a href="https://docs.python.org/3/library/exceptions.html">Built-in Exceptions</a> docs).\n<!-- </tip> -->\n</td></tr></tbody></table>\n\n<hr>\n</details>\n\n## Contributing\n\nAny and all contributions are absolutely welcome. Feel free to open an issue or just jump straight to a PR. Let\'s discuss and make this the best it can be! 😄\n\n## License\n\nLethargy is released under the [MIT license](https://github.com/SeparateRecords/lethargy/blob/master/LICENSE).\n',
    'author': 'SeparateRecords',
    'author_email': 'me@rob.ac',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/SeparateRecords/lethargy',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
